/**************************************************************************/

/* Routines for extending and reducing a triangulation.
   General principle:  extendx(e); reducex(e)  will extend by one
   vertex of degree x (x=3,4,n) then reduce it to the original graph.
   The final graph is exactly the same as the original
   (including pointer values) except that possibly the values of
   firstedge[] might be different.
*/

static void
extend3(EDGE *e)

/* inserts a vertex with valence 3 in the triangle on the right hand
   side (->next direction) of the edge e */
{
  register EDGE *work1, *work2, *work3;
  register EDGE *oe, *owork1, *owork2, *owork3;
  int v0,v1,v2,i;

  work1 = STAR3(nv);
  work2 = work1+1;
  work3 = work2+1;
  owork1 = work1+NUMEDGES;
  owork2 = owork1+1;
  owork3 = owork2+1;
  firstedge[nv] = work3+1;

  /* work1 starts at the beginning of e: */

  work1->start = work1->invers->end = e->start;
  e->next->prev = work1;
  work1->next = e->next;
  work1->prev = e;
  e->next = work1;
  oe = e->op;
  owork1->start = owork1->invers->end = oe->start;
  oe->prev->next = owork1;
  owork1->prev = oe->prev;
  owork1->next = oe;
  oe->prev = owork1;
  (degree[e->start])++;

  adj[e->start][nv] = adj[nv][e->start] = TRUE;
  v0 = e->start;

  /* Now go one edge further around the triangle and the same once more */

  e = e->invers->prev;

  work2->start = work2->invers->end = e->start;
  e->next->prev = work2;
  work2->next = e->next;
  work2->prev = e;
  e->next = work2;
  oe = e->op;
  owork2->start = owork2->invers->end = oe->start;
  oe->prev->next = owork2;
  owork2->prev = oe->prev;
  owork2->next = oe;
  oe->prev = owork2;
  (degree[e->start])++;

  adj[e->start][nv] = adj[nv][e->start] = TRUE;
  v1 = e->start;

  e = e->invers->prev;

  work3->start = work3->invers->end = e->start;
  e->next->prev = work3;
  work3->next = e->next;
  work3->prev = e;
  e->next = work3;
  oe = e->op;
  owork3->start = owork3->invers->end = oe->start;
  oe->prev->next = owork3;
  owork3->prev = oe->prev;
  owork3->next = oe;
  oe->prev = owork3;
  (degree[e->start])++;

  adj[e->start][nv] = adj[nv][e->start] = TRUE;
  v2 = e->start;

  degree[nv] = 3;

  for (i=0;i<nv;i++)
    nnei[i][nv] = nnei[nv][i] = adj[v0][i] + adj[v1][i] + adj[v2][i];

  if (nnei[v0][v1]==2)
    {
      ncontract[v0]--;
      ncontract[v1]--;
    }
  if (nnei[v1][v2]==2)
    {
      ncontract[v1]--;
      ncontract[v2]--;
    }
  if (nnei[v2][v0]==2)
    {
      ncontract[v2]--;
      ncontract[v0]--;
    }
  ncontract[v0]++;
  ncontract[v1]++;
  ncontract[v2]++;
  ncontract[nv] = 3;

  nnei[v0][v1] = ++nnei[v1][v0];
  nnei[v1][v2] = ++nnei[v2][v1];
  nnei[v2][v0] = ++nnei[v0][v2];

  /* Now I have 6 edges and one vertex more */

  ne += 6;
  nv++;

  check_it(903);
}

/****************************************/

static void
reduce3(EDGE *e)

/* deletes a vertex with valence 3 in the triangle on the right hand side
 (->next-direction) of the edge e. It is not checked whether the vertex
 really has valence 3 -- this has to be made sure in advance      */
{
  int v0,v1,v2;
  EDGE *oe;

  /* It might be that one of the edges leading to the new vertex now is
     an entry of firstedge[] */

  /* If (firstedge[e->start]==e->next) would take too much time, so just*/

  firstedge[e->start] = e;
  
  e->next = e->next->next; e->next->prev = e;
  oe = e->op;
  oe->prev = oe->prev->prev; oe->prev->next = oe;
  (degree[e->start])--;

  adj[e->start][nv-1] = adj[nv-1][e->start] = FALSE;
  v0 = e->start;

  e = e->invers;

  firstedge[e->start] = e;
  /* Now delete on the ->prev side */
  e->prev = e->prev->prev; e->prev->next = e;
  oe = e->op;
  oe->next = oe->next->next; oe->next->prev = oe;
  (degree[e->start])--;

  adj[e->start][nv-1] = adj[nv-1][e->start] = FALSE;
  v1 = e->start;

  e = e->prev->invers;

  firstedge[e->start] = e;
  /* Again on the ->prev side: */
  e->prev = e->prev->prev; e->prev->next = e;
  oe = e->op;
  oe->next = oe->next->next; oe->next->prev = oe;
  (degree[e->start])--;

  adj[e->start][nv-1] = adj[nv-1][e->start] = FALSE;
  v2 = e->start;

  nnei[v0][v1] = --nnei[v1][v0];
  nnei[v1][v2] = --nnei[v2][v1];
  nnei[v2][v0] = --nnei[v0][v2];

  ncontract[v0]--;
  ncontract[v1]--;
  ncontract[v2]--;
  if (nnei[v0][v1]==2)
    {
      ncontract[v0]++;
      ncontract[v1]++;
    }
  if (nnei[v1][v2]==2)
    {
      ncontract[v1]++;
      ncontract[v2]++;
    }
  if (nnei[v2][v0]==2)
    {
      ncontract[v2]++;
      ncontract[v0]++;
    }

  nv--;
  ne -= 6;

  check_it(993);
}

/************************************************************************/

static void
extend4(EDGE *e, EDGE *list[])

/* Deletes e->next and its inverse and puts a valence 4 vertex into the
   resulting square.
   In list[0..1] the deleted edges are stored. This list must be handed
   to reduce4() */
{
  register EDGE *work1, *work2, *work3, *work4;
  register EDGE *oe, *owork1, *owork2, *owork3, *owork4;
  int v0,v1,v2,v3,i,k,v;

  list[0] = e->next; list[1] = e->next->invers;

  work1 = STAR4(nv);
  work2 = work1+1;
  work3 = work2+1;
  work4 = work3+1;
  owork1 = work1+NUMEDGES;
  owork2 = owork1+1;
  owork3 = owork2+1;
  owork4 = owork3+1;
  firstedge[nv] = work4+1;

  /* make sure firstedge points at a valid edge afterwards */

  firstedge[e->start] = e;

  /* work1 starts at the beginning of e: */

  work1->start = work1->invers->end = e->start;
  work1->next = e->next->next;
  work1->next->prev = work1;
  work1->prev = e;
  e->next = work1;
  oe = e->op;
  owork1->start = owork1->invers->end = oe->start;
  owork1->prev = oe->prev->prev;
  owork1->prev->next = owork1;
  owork1->next = oe;
  oe->prev = owork1;
  /* the degree of e->start doesn't change */

  adj[e->start][nv] = adj[nv][e->start] = TRUE;
  v0 = e->start;

  /* Now go one edge further around the square: */

  e = e->invers->prev;

  work2->start = work2->invers->end = e->start;
  e->next->prev = work2;
  work2->next = e->next;
  work2->prev = e;
  e->next = work2;
  oe = e->op;
  owork2->start = owork2->invers->end = oe->start;
  oe->prev->next = owork2;
  owork2->prev = oe->prev;
  owork2->next = oe;
  oe->prev = owork2;
  (degree[e->start])++;

  adj[e->start][nv] = adj[nv][e->start] = TRUE;
  v1 = e->start;

  /* Now we have one edge to jump about again: */
  e = e->invers->prev->prev;

  firstedge[e->start] = e;
  /* Again an edge is deleted... */

  work3->start = work3->invers->end = e->start;
  work3->next = e->next->next;
  work3->next->prev = work3;
  work3->prev = e;
  e->next = work3;
  oe = e->op;
  owork3->start = owork3->invers->end = oe->start;
  owork3->prev = oe->prev->prev;
  owork3->prev->next = owork3;
  owork3->next = oe;
  oe->prev = owork3;
  /* the degree of e->start doesn't change */

  adj[e->start][nv] = adj[nv][e->start] = TRUE;
  v2 = e->start;

  /* Now go again one edge further around the square: */

  e = e->invers->prev;

  work4->start = work4->invers->end = e->start;
  e->next->prev = work4;
  work4->next = e->next;
  work4->prev = e;
  e->next = work4;
  oe = e->op;
  owork4->start = owork4->invers->end = oe->start;
  oe->prev->next = owork4;
  owork4->prev = oe->prev;
  owork4->next = oe;
  oe->prev = owork4;
  (degree[e->start])++;

  adj[e->start][nv] = adj[nv][e->start] = TRUE;
  v3 = e->start;

  degree[nv] = 4;

  adj[v0][v2] = adj[v2][v0] = FALSE;

  for (i=0;i<nv;i++)
    nnei[i][nv] = nnei[nv][i] =
      adj[v0][i] + adj[v1][i] + adj[v2][i] + adj[v3][i];

  if (nnei[v0][v2] > 2)
    {
      ncontract[v0]++;
      ncontract[v2]++;
    }

  if (adj[v1][v3])
    ncontract[nv] = 2;
  else
    {
      ncontract[nv] = 4;
      ncontract[v1]++;
      ncontract[v3]++;
    }

  nnei[v0][v2] = ++nnei[v2][v0];

  nnei[v1][v3] = ++nnei[v3][v1];

  e = list[0]->next;
  for (k=0;k<degree[v0]-3;k++)
    {
      e = e->next;
      v = e->end;
      nnei[v2][v] = --nnei[v][v2];
      if (adj[v2][v] && nnei[v2][v]==2)
	{
	  ncontract[v]++;
	  ncontract[v2]++;
	}
    }

  e = list[1]->next;
  for (k=0;k<degree[v2]-3;k++)
    {
      e = e->next;
      v = e->end;
      nnei[v0][v] = --nnei[v][v0];
      if (adj[v0][v] && nnei[v0][v]==2)
	{
	  ncontract[v]++;
	  ncontract[v0]++;
	}
    }

  /* Now I have 6 edges and one vertex more */

  ne += 6;
  nv++;

  check_it(904);
}

/**************************/

static void
reduce4(EDGE *e, EDGE *list[])

/* The inverse operation to extend4().
   Deletes the vertex with valence 4 in the triangle on the right hand side
   (->next-direction) of the edge e that is not contained in e. It is not
   checked whether the vertex really has valence 4 -- this has to be made
   sure in advance. The vector list[] must contain the edges deleted before.
   It might be that one of the edges leading to the new vertex now is
   an entry of firstedge[] */
{
  int v0,v1,v2,v3,i,v;

  firstedge[e->start] = e;

  list[0]->next->prev = list[0];
  e->next = list[0];
  list[0]->op->prev->next = list[0]->op;
  e->op->prev = list[0]->op;

  v0 = e->start;
  adj[v0][nv-1] = adj[nv-1][v0] = FALSE;

  e = e->invers;

  firstedge[e->start] = e;
  /* Now delete on the ->prev side */
  e->prev = e->prev->prev; e->prev->next = e;
  e->op->next = e->op->next->next; e->op->next->prev = e->op;
  (degree[e->start])--;

  adj[e->start][nv-1] = adj[nv-1][e->start] = FALSE;
  v1 = e->start;

  e = e->prev->invers;

  firstedge[e->start] = e;
  /* Again on the ->prev side: */
  list[1]->prev->next = list[1];
  e->prev = list[1];
  list[1]->op->next->prev = list[1]->op;
  e->op->next = list[1]->op;

  adj[e->start][nv-1] = adj[nv-1][e->start] = FALSE;
  v2 = e->start;

  e = list[1]->prev->invers;
  firstedge[e->start] = e;
  e->prev = e->prev->prev; e->prev->next = e;
  e->op->next = e->op->next->next; e->op->next->prev = e->op;
  (degree[e->start])--;

  adj[e->start][nv-1] = adj[nv-1][e->start] = FALSE;
  v3 = e->start;

  e = list[0]->next;
  for (i=0;i<degree[v0]-3;i++)
    {
      e = e->next;
      v = e->end;
      nnei[v2][v] = ++nnei[v][v2];
      if (adj[v2][v] && nnei[v2][v]==3)
	{	
	  ncontract[v]--;
	  ncontract[v2]--;
	}
    }
  e = list[1]->next;
  for (i=0;i<degree[v2]-3;i++)
    {
      e = e->next;
      v = e->end;
      nnei[v0][v] = ++nnei[v][v0];
      if (adj[v0][v] && nnei[v0][v]==3)
	{
	  ncontract[v]--;
	  ncontract[v0]--;
	}
    }

  nnei[v0][v2] = --nnei[v2][v0];
  nnei[v1][v3] = --nnei[v3][v1];

  if (nnei[v0][v2] > 2)
    {
      ncontract[v0]--;
      ncontract[v2]--;
    }

  if (!adj[v1][v3])
    {
      ncontract[v1]--;
      ncontract[v3]--;
    }

  adj[v0][v2] = adj[v2][v0] = TRUE;

  nv--;
  ne -= 6;

  check_it(994);
}

/**********************************************************************/

static void
extendn(int n, EDGE *e, EDGE *list[])

/* Deletes (n-3) edges in the next direction (e->next, e->next->next, ...)
   and their inverse and puts a valence
   n vertex into the resulting polygon.
   In list[0..(n-3)*2-1] the deleted edges are stored.
   This list must be handed to reducen() */
{
  register EDGE *work, *e1;
  int v[MAXN],i,j,k,vend;
  EDGE *oe, *owork;

  work = e;
  for (j=0;j<n-3;j++)
    {
      work = work->next;
      list[2*j] = work;
      list[2*j+1] = work->invers;
    }

  work = STARN(nv);
  firstedge[nv] = work+MAXN;
  firstedge[nv]->prev = work+MAXN+n-1;
  firstedge[nv]->prev->next = firstedge[nv];
  owork = firstedge[nv]->op;
  owork->next = firstedge[nv]->prev->op;
  owork->next->prev = owork;

  /* make sure firstedge points at a valid edge afterwards */

  firstedge[e->start] = e;

  /* work starts at the beginning of e: */

  work->start = work->invers->end = e->start;
  e1 = e;
  for (j=0;j<n-2;j++)
    e1 = e1->next;
  work->next = e1;
  work->next->prev = work;
  work->prev = e;
  e->next = work;
  oe = e->op;
  owork = work->op;
  owork->start = owork->invers->end = oe->start;
  owork->prev = e1->op;
  owork->prev->next = owork;
  owork->next = oe;
  oe->prev = owork;
  degree[e->start] = degree[e->start] - (n-4);

  v[0] = e->start;

  /* Now go one edge further around the pentagon: */
  e = e->invers->prev;

  work = work+1;
  work->start = work->invers->end = e->start;
  e->next->prev = work;
  work->next = e->next;
  work->prev = e;
  e->next = work;
  oe = e->op;
  owork = work->op;
  owork->start = owork->invers->end = oe->start;
  oe->prev->next = owork;
  owork->prev = oe->prev;
  owork->next = oe;
  oe->prev = owork;
  (degree[e->start])++;

  v[1] = e->start;

  /* Now go one edge further around the pentagon jumping over one edge: */

  for (j=2;j<n-1;j++)
    {
      e = e->invers->prev->prev;

      firstedge[e->start] = e;

      /* here also an edge is deleted */

      work = work+1;
      work->start = work->invers->end = e->start;
      work->next = e->next->next;
      work->next->prev = work;
      work->prev = e;
      e->next = work;
      oe = e->op;
      owork = work->op;
      owork->start = owork->invers->end = oe->start;
      owork->prev = oe->prev->prev;
      owork->prev->next = owork;
      owork->next = oe;
      oe->prev = owork;
      /* the degree of e->start doesn't change */

      v[j] = e->start;
    }

  /* Finally go one edge further around the pentagon: */
  e = e->invers->prev;

  work = work+1;
  work->start = work->invers->end = e->start;
  e->next->prev = work;
  work->next = e->next;
  work->prev = e;
  e->next = work;
  oe = e->op;
  owork = work->op;
  owork->start = owork->invers->end = oe->start;
  oe->prev->next = owork;
  owork->prev = oe->prev;
  owork->next = oe;
  oe->prev = owork;
  (degree[e->start])++;

  v[n-1] = e->start;

  degree[nv] = n;

  /* do bookkeeping for removing edges (diagonals) */

  /* remove each diagonal edge.
     if nnei of its ends was 2 then it was contractible so reduce
     ncontract for its ends. */

  for (j=2;j<n-1;j++)
    {
      if (nnei[v[0]][v[j]] == 2)
	{
	  ncontract[v[0]]--;
	  ncontract[v[j]]--;
	}
      adj[v[0]][v[j]] = adj[v[j]][v[0]] = FALSE;
    }

  /* for each neighbor, vend, of v[0] reduce nnei of vend and
     each other end (v[j],j=2,..,n-2) of the diagonals.
     if vend and v[j] are adjacent and nnei is now 2 then increase
     ncontract for vend and v[j].
     don't bother with vend=(v[i],i=1,..,n-1) since nnei[v[i]][v[j]]
     would then need to be increased later
     when radial edges are added. (iou A) */

  e = list[2*(n-4)]->next;
  for (k=0;k<degree[v[0]]-3;k++)
    {
      e = e->next;
      vend = e->end;
      for (j=2;j<n-1;j++)
	{
	  nnei[v[j]][vend] = --nnei[vend][v[j]];
	  if (adj[v[j]][vend] && nnei[v[j]][vend]==2)
	    {
	      ncontract[vend]++;
	      ncontract[v[j]]++;
	    }
	}
    }

  /* for each neighbor, vend, of each non-v[0] end of the diagonals
     (v[j],j=2,..,n-2) reduce nnei of vend and v[0].
     if vend and v[0] are adjacent and nnei is now 2 then increase
     ncontract for vend and v[0].
     don't bother with edges on path around v[nv] will deal with them later
     (iou B) */

  for (j=2;j<n-1;j++)
    {
      e = list[2*j-3]->next;
      for (k=0;k<degree[v[j]]-3;k++)
	{
	  e = e->next;
	  vend = e->end;
	  nnei[v[0]][vend] = --nnei[vend][v[0]];
	  if (adj[v[0]][vend] && nnei[v[0]][vend]==2)
	    {
	      ncontract[vend]++;
	      ncontract[v[0]]++;
	    }
	}
    }

  /* do bookkeeping for adding edges (radial from new vertex) */

  /* for each old vertex, i, compute nnei[nv][i] by counting number of new
     neighbors of nv (v[j],j=0,..,n-1) which are adjacent to i. */

  for (i=0;i<nv;i++)
    {
      nnei[nv][i] = 0;
      for (j=0;j<n;j++)
	nnei[nv][i] += adj[v[j]][i];
      nnei[i][nv] = nnei[nv][i];
    }

  /* for each new edge if nnei is 2 increase ncontract of ends */

  ncontract[nv] = 0;
  for (j=0;j<n;j++)
    {
      adj[v[j]][nv] = adj[nv][v[j]] = TRUE;
      if (nnei[v[j]][nv] == 2)
	{
	  ncontract[v[j]]++;
	  ncontract[nv]++;
	}
    }

  /* update nnei[v[i]][v[j]] because of addition of nv.
     would need to increase for each pair (((i,j),i=0,..,n-2),j=i+1,..,n-1).
     from iou A above we didn't decrease (((i,j),i=1,n-2),j=i+1,..,n-1)
     except (1,n-1).
     from iou B above we didn't decrease (((0,j),i=2,..,n-2),j=i-1,i+1).
     so still need to decrease ((0,j),j=3,..,n-3) and  increase (1,n-1).
  */

  /* decrease ((0,j),j=3,..,n-3). */

  for (j=3;j<n-2;j++)
    nnei[v[0]][v[j]] = --nnei[v[j]][v[0]];

  /* increase (1,n-1). */

  nnei[v[1]][v[n-1]] = ++nnei[v[n-1]][v[1]];
  if (adj[v[1]][v[n-1]] && nnei[v[1]][v[n-1]]==3)
    {
      ncontract[v[1]]--;
      ncontract[v[n-1]]--;
    }

  /* Now I have 6 edges and one vertex more */

  ne += 6;
  nv++;

  check_it(900+n);
}

/*****************************/

static void
reducen(int n, EDGE *e, EDGE *list[])

/* The inverse operation to extendn().
   Deletes the vertex with valence n at the end of e->next. It is not
   checked whether the vertex really has valence n -- this has to be made
   sure in advance. The vector list[] must contain the edges deleted before.
   It might be that one of the edges leading to the new vertex now is
   an entry of firstedge[] */
{
  int v[MAXN],j,k,vend;
  EDGE *work;

  firstedge[e->start] = e;

  e->next = list[0];
  list[2*(n-4)]->next->prev = list[2*(n-4)];
  e->op->prev = list[0]->op;
  list[2*(n-4)]->op->prev->next = list[2*(n-4)]->op;
  degree[e->start] = degree[e->start] + (n-4);

  v[0] = e->start;

  e = e->invers;

  /* Delete the edge on the prev side */
  firstedge[e->start] = e;
  e->prev = e->prev->prev; e->prev->next = e;
  e->op->next = e->op->next->next; e->op->next->prev = e->op;
  (degree[e->start])--;

  v[1] = e->start;

  e = e->prev->invers;

  for (j=2;j<n-1;j++)
    {
      firstedge[e->start] = e;
      /* Delete the edge and insert the old edge: */
      e->prev = list[2*j-3];
      list[2*j-3]->prev->next = list[2*j-3];
      e->op->next = list[2*j-3]->op;
      list[2*j-3]->op->next->prev = list[2*j-3]->op;

      v[j] = e->start;

      e = e->prev->prev->invers;
    }

  firstedge[e->start] = e;
  e->prev = e->prev->prev; e->prev->next = e;
  e->op->next = e->op->next->next; e->op->next->prev = e->op;
  (degree[e->start])--;

  v[n-1] = e->start;

  work = STARN(nv-1)+MAXN;
  work->prev->next = STARN(nv-1) + MAXN + n;
  work->op->next->prev = work->prev->next->op;
  work->prev = STARN(nv-1) + MAXN + MAXN - 1;
  work->op->next = work->prev->op;

  /* do reverse operations as in extendn.
     see comments in extendn. */

  /* do bookkeeping for removing edges (radial from new vertex) */

  for (j=0;j<n;j++)
    {
      adj[v[j]][nv-1] = adj[nv-1][v[j]] = FALSE;
      if (nnei[v[j]][nv-1] == 2)
	ncontract[v[j]]--;
    }

  for (j=3;j<n-2;j++)
    nnei[v[0]][v[j]] = ++nnei[v[j]][v[0]];

  nnei[v[1]][v[n-1]] = --nnei[v[n-1]][v[1]];

  if (adj[v[1]][v[n-1]] && nnei[v[1]][v[n-1]] == 2)
    {
      ncontract[v[1]]++;
      ncontract[v[n-1]]++;
    }

  /* do bookkeeping for adding edges (diagonals) */

  e = list[2*(n-4)]->next;
  for (k=0;k<degree[v[0]]-(n-1);k++)
    {
      e = e->next;
      vend = e->end;
      for (j=2;j<n-1;j++)
	{
	  nnei[v[j]][vend] = ++nnei[vend][v[j]];
	  if (adj[v[j]][vend] && nnei[v[j]][vend]==3)
	    {
	      ncontract[vend]--;
	      ncontract[v[j]]--;
	    }
	}
    }

  for (j=2;j<n-1;j++)
    {
      e = list[2*j-3]->next;
      for (k=0;k<degree[v[j]]-3;k++)
	{
	  e = e->next;
	  vend = e->end;
	  nnei[v[0]][vend] = ++nnei[vend][v[0]];
	  if (adj[v[0]][vend] && nnei[v[0]][vend]==3)
	    {
	      ncontract[vend]--;
	      ncontract[v[0]]--;
	    }
	}
    }

  for (j=2;j<n-1;j++)
    {
      if (nnei[v[0]][v[j]] == 2)
	{
	  ncontract[v[0]]++;
	  ncontract[v[j]]++;
	}
      adj[v[0]][v[j]] = adj[v[j]][v[0]] = TRUE;
    }

  nv--;
  ne-=6;

  check_it(990+n);
}

/**************************************************************************/

/* Routines for extending and reducing a triangulation with holes.
   General principle:  extendx(e); reducex(e)  will extend by one
   vertex of degree x (x=3,4,n) then reduce it to the original graph.
   The final graph is exactly the same as the original
   (including pointer values) except that possibly the values of
   firstedge[] might be different.
*/

static void
extend3_holes(EDGE *e)

/* inserts a vertex with valence 3 in the triangle on the right hand
   side (->next direction) of the edge e */
{
  register EDGE *work1, *work2, *work3;
  register EDGE *oe, *owork1, *owork2, *owork3;
  EDGE *e3;
  int v0,v1,v2,i;
  int hole;

  work1 = STAR3(nv);
  work2 = work1+1;
  work3 = work2+1;
  owork1 = work1+NUMEDGES;
  owork2 = owork1+1;
  owork3 = owork2+1;
  firstedge[nv] = work3+1;

  /* in case not a triangulation e->next now */

  e3 = e->next;
  hole = RIGHTHOLE(e);

  /* work1 starts at the beginning of e: */

  work1->start = work1->invers->end = e->start;
  e->next->prev = work1;
  work1->next = e->next;
  work1->prev = e;
  e->next = work1;
  oe = e->op;
  owork1->start = owork1->invers->end = oe->start;
  oe->prev->next = owork1;
  owork1->prev = oe->prev;
  owork1->next = oe;
  oe->prev = owork1;
  (degree[e->start])++;

  adj[e->start][nv] = adj[nv][e->start] = TRUE;
  v0 = e->start;

  /* Now go one edge further around the triangle and the same once more */

  e = e->invers->prev;

  work2->start = work2->invers->end = e->start;
  e->next->prev = work2;
  work2->next = e->next;
  work2->prev = e;
  e->next = work2;
  oe = e->op;
  owork2->start = owork2->invers->end = oe->start;
  oe->prev->next = owork2;
  owork2->prev = oe->prev;
  owork2->next = oe;
  oe->prev = owork2;
  (degree[e->start])++;

  adj[e->start][nv] = adj[nv][e->start] = TRUE;
  v1 = e->start;

  e = e3->invers;

  work3->start = work3->invers->end = e->start;
  e->next->prev = work3;
  work3->next = e->next;
  work3->prev = e;
  e->next = work3;
  oe = e->op;
  owork3->start = owork3->invers->end = oe->start;
  oe->prev->next = owork3;
  owork3->prev = oe->prev;
  owork3->next = oe;
  oe->prev = owork3;
  (degree[e->start])++;

  adj[e->start][nv] = adj[nv][e->start] = TRUE;
  v2 = e->start;

  degree[nv] = 3;

  for (i=0;i<nv;i++)
    nnei[i][nv] = nnei[nv][i] = adj[v0][i] + adj[v1][i] + adj[v2][i];
  if (nnei[v0][v1]==2)
    {
      ncontract[v0]--;
      ncontract[v1]--;
    }
  if (nnei[v1][v2]==2)
    {
      ncontract[v1]--;
      ncontract[v2]--;
    }
  if (nnei[v2][v0]==2)
    {
      ncontract[v2]--;
      ncontract[v0]--;
    }
  if (!hole) 
    {
      ncontract[v0]++;
      ncontract[v1]++;
      ncontract[v2]++;
      ncontract[nv] = 3;
    }
  else
    {
      ncontract[v0]++;
      ncontract[nv] = 1;
      if (nnei[nv][v1] == 1) // ???
	{
	  ncontract[v1]++;
	  ncontract[nv]++;
	}
      if (nnei[nv][v2] == 1) // ???
	{
	  ncontract[v2]++;
	  ncontract[nv]++;
	}
    }

  if (hole)
    {
      nnei[v1][nv] = ++nnei[nv][v1];
      nnei[v2][nv] = ++nnei[nv][v2];
    }
  else
    {
      nnei[v0][v1] = ++nnei[v1][v0];
      nnei[v2][v0] = ++nnei[v0][v2];
    }
  nnei[v1][v2] = ++nnei[v2][v1];

  /* Now I have 6 edges and one vertex more */

  ne += 6;
  nv++;

  
  check_it(303);
  
}

/****************************************/

static void
reduce3_holes(EDGE *e)

/* deletes a vertex with valence 3 in the triangle on the right hand side
 (->next-direction) of the edge e. It is not checked whether the vertex
 really has valence 3 -- this has to be made sure in advance      */
{
  int v0,v1,v2;
  EDGE *oe;
  int hole;

  hole = RIGHTHOLE(e->next->invers->next);

  /* It might be that one of the edges leading to the new vertex now is
     an entry of firstedge[] */

  /* If (firstedge[e->start]==e->next) would take too much time, so just*/

  firstedge[e->start] = e;
  e->next = e->next->next; e->next->prev = e;
  oe = e->op;
  oe->prev = oe->prev->prev; oe->prev->next = oe;
  (degree[e->start])--;

  adj[e->start][nv-1] = adj[nv-1][e->start] = FALSE;
  v0 = e->start;

  e = e->invers;

  firstedge[e->start] = e;
  /* Now delete on the ->prev side */
  e->prev = e->prev->prev; e->prev->next = e;
  oe = e->op;
  oe->next = oe->next->next; oe->next->prev = oe;
  (degree[e->start])--;

  adj[e->start][nv-1] = adj[nv-1][e->start] = FALSE;
  v1 = e->start;

  e = e->invers->next->invers->next->next;

  firstedge[e->start] = e;
  /* Again on the ->prev side: */
  e->prev = e->prev->prev; e->prev->next = e;
  oe = e->op;
  oe->next = oe->next->next; oe->next->prev = oe;
  (degree[e->start])--;

  adj[e->start][nv-1] = adj[nv-1][e->start] = FALSE;
  v2 = e->start;

  if (!hole) 
    {
      nnei[v0][v1] = --nnei[v1][v0];
      nnei[v2][v0] = --nnei[v0][v2];
    }
  nnei[v1][v2] = --nnei[v2][v1];

  // ??

  //  if (adj[v1][v2]) 
  if (!hole)
    {
      ncontract[v0]--;
      ncontract[v1]--;
      ncontract[v2]--;
    }
  else
    {
      ncontract[v0]--;
      if (nnei[nv-1][v1] == 2)
	ncontract[v1]--;
      if (nnei[nv-1][v2] == 2)
	ncontract[v2]--;
    }
  if (nnei[v0][v1]==2)
    {
      ncontract[v0]++;
      ncontract[v1]++;
    }
  if (nnei[v1][v2]==2 && adj[v1][v2])
    {
      ncontract[v1]++;
      ncontract[v2]++;
    }
  if (nnei[v2][v0]==2)
    {
      ncontract[v2]++;
      ncontract[v0]++;
    }

  nv--;
  ne -= 6;

  
  check_it(393);
  
}

/************************************************************************/

static void
extend4_holes(EDGE *e, EDGE *list[])

/* Deletes e->next and its inverse and puts a valence 4 vertex into the
   resulting square.
   In list[0..1] the deleted edges are stored. This list must be handed
   to reduce4() */
{
  register EDGE *work1, *work2, *work3, *work4;
  register EDGE *oe, *owork1, *owork2, *owork3, *owork4;
  int v0,v1,v2,v3,i,k,v;
  int hole1,hole2;

  hole1 = RIGHTHOLE(e);
  hole2 = RIGHTHOLE(e->next);

  list[0] = e->next; list[1] = e->next->invers;

  work1 = STAR4(nv);
  work2 = work1+1;
  work3 = work2+1;
  work4 = work3+1;
  owork1 = work1+NUMEDGES;
  owork2 = owork1+1;
  owork3 = owork2+1;
  owork4 = owork3+1;
  firstedge[nv] = work4+1;

  firstedge[e->start] = e;
  /* make sure firstedge points at a valid edge afterwards */

  /* work1 starts at the beginning of e: */

  work1->start = work1->invers->end = e->start;
  work1->next = e->next->next;
  work1->next->prev = work1;
  work1->prev = e;
  e->next = work1;
  oe = e->op;
  owork1->start = owork1->invers->end = oe->start;
  owork1->prev = oe->prev->prev;
  owork1->prev->next = owork1;
  owork1->next = oe;
  oe->prev = owork1;
  /* the degree of e->start doesn't change */

  adj[e->start][nv] = adj[nv][e->start] = TRUE;
  v0 = e->start;

  /* Now go one edge further around the square: */

  e = e->invers->prev;

  work2->start = work2->invers->end = e->start;
  e->next->prev = work2;
  work2->next = e->next;
  work2->prev = e;
  e->next = work2;
  oe = e->op;
  owork2->start = owork2->invers->end = oe->start;
  oe->prev->next = owork2;
  owork2->prev = oe->prev;
  owork2->next = oe;
  oe->prev = owork2;
  (degree[e->start])++;

  adj[e->start][nv] = adj[nv][e->start] = TRUE;
  v1 = e->start;

  /* Now we have one edge to jump about again: */
  
  e = list[1]->prev;

  firstedge[e->start] = e;
  /* Again an edge is deleted... */

  work3->start = work3->invers->end = e->start;
  work3->next = e->next->next;
  work3->next->prev = work3;
  work3->prev = e;
  e->next = work3;
  oe = e->op;
  owork3->start = owork3->invers->end = oe->start;
  owork3->prev = oe->prev->prev;
  owork3->prev->next = owork3;
  owork3->next = oe;
  oe->prev = owork3;
  /* the degree of e->start doesn't change */

  adj[e->start][nv] = adj[nv][e->start] = TRUE;
  v2 = e->start;

  /* Now go again one edge further around the square: */

  e = list[0]->next->invers;

  work4->start = work4->invers->end = e->start;
  e->next->prev = work4;
  work4->next = e->next;
  work4->prev = e;
  e->next = work4;
  oe = e->op;
  owork4->start = owork4->invers->end = oe->start;
  oe->prev->next = owork4;
  owork4->prev = oe->prev;
  owork4->next = oe;
  oe->prev = owork4;
  (degree[e->start])++;

  adj[e->start][nv] = adj[nv][e->start] = TRUE;
  v3 = e->start;

  degree[nv] = 4;

  adj[v0][v2] = adj[v2][v0] = FALSE;

  if (nnei[v0][v2] > 2)
    {
      ncontract[v0]++;
      ncontract[v2]++;
    }

  for (i=0;i<nv;i++)
    nnei[i][nv] = nnei[nv][i] =
      adj[v0][i] + adj[v1][i] + adj[v2][i] + adj[v3][i];
  if (hole1)
    {
      nnei[v1][nv] = ++nnei[nv][v1]; 
      nnei[v2][nv] = ++nnei[nv][v2]; 
      //nnei[v0][v1] = --nnei[v1][v0];
      nnei[v0][v2] = --nnei[v2][v0];
    }
  if (hole2)
    {
      nnei[v2][nv] = ++nnei[nv][v2]; 
      nnei[v3][nv] = ++nnei[nv][v3];
      //nnei[v0][v3] = --nnei[v3][v0];
      nnei[v0][v2] = --nnei[v2][v0];
    }

  if (nnei[nv][v2] > 2)
    ncontract[v2]--;
  
  ncontract[nv] = 1 + (nnei[v1][nv] == 2) + (nnei[v2][nv] == 2) + (nnei[v3][nv] == 2);
  
  ncontract[v1] += (nnei[v1][nv] == 2);
  ncontract[v3] += (nnei[v3][nv] == 2);
  
  nnei[v0][v2] = ++nnei[v2][v0];

  nnei[v1][v3] = ++nnei[v3][v1];
  if (adj[v1][v3] && nnei[v1][v3]==3)
    {
      ncontract[v1]--;
      ncontract[v3]--;
    }

  e = list[0]->next;
  for (k=0;k<degree[v0]-3;k++)
    {
      e = e->next;
      v = e->end;
      nnei[v2][v] = --nnei[v][v2];
      if (adj[v2][v] && nnei[v2][v]==2)
	{
	  ncontract[v]++;
	  ncontract[v2]++;
	}
    }

  e = list[1];
  if (!hole1) e = e->next;
  for (k=0;k<degree[v2]-3+hole1+hole2;k++)
    {
      e = e->next;
      v = e->end;
      nnei[v0][v] = --nnei[v][v0];
      if (adj[v0][v] && nnei[v0][v]==2)
	{
	  ncontract[v]++;
	  ncontract[v0]++;
	}
    }

  /* Now I have 6 edges and one vertex more */

  ne += 6;
  nv++;

  
  check_it(304);
  
}

/**************************/

static void
reduce4_holes(EDGE *e, EDGE *list[])

/* The inverse operation to extend4().
   Deletes the vertex with valence 4 in the triangle on the right hand side
   (->next-direction) of the edge e that is not contained in e. It is not
   checked whether the vertex really has valence 4 -- this has to be made
   sure in advance. The vector list[] must contain the edges deleted before.
   It might be that one of the edges leading to the new vertex now is
   an entry of firstedge[] */
{
  int v0,v1,v2,v3,i,v;
  int hole1,hole2;

  firstedge[e->start] = e;

  list[0]->next->prev = list[0];
  e->next = list[0];
  list[0]->op->prev->next = list[0]->op;
  e->op->prev = list[0]->op;

  v0 = e->start;
  adj[v0][nv-1] = adj[nv-1][v0] = FALSE;

  e = e->invers;

  firstedge[e->start] = e;
  /* Now delete on the ->prev side */
  e->prev = e->prev->prev; e->prev->next = e;
  e->op->next = e->op->next->next; e->op->next->prev = e->op;
  (degree[e->start])--;

  adj[e->start][nv-1] = adj[nv-1][e->start] = FALSE;
  v1 = e->start;

  e = list[1]->next;

  firstedge[e->start] = e;
  /* Again on the ->prev side: */
  list[1]->prev->next = list[1];
  e->prev = list[1];
  list[1]->op->next->prev = list[1]->op;
  e->op->next = list[1]->op;

  adj[e->start][nv-1] = adj[nv-1][e->start] = FALSE;
  v2 = e->start;

  e = list[0]->next->invers->next->next;
  firstedge[e->start] = e;
  e->prev = e->prev->prev; e->prev->next = e;
  e->op->next = e->op->next->next; e->op->next->prev = e->op;
  (degree[e->start])--;

  adj[e->start][nv-1] = adj[nv-1][e->start] = FALSE;
  v3 = e->start;

  hole1 = RIGHTHOLE(list[1]);
  hole2 = RIGHTHOLE(list[0]);

  e = list[0]->next;
  for (i=0;i<degree[v0]-3;i++)
    {
      e = e->next;
      v = e->end;
      nnei[v2][v] = ++nnei[v][v2];
      if (adj[v2][v] && nnei[v2][v]==3)
	{	
	  ncontract[v]--;
	  ncontract[v2]--;
	}
    }
  e = list[1];
  if (!hole1) e = e->next;
  for (i=0;i<degree[v2]-3+hole1+hole2;i++)
    {
      e = e->next;
      v = e->end;
      nnei[v0][v] = ++nnei[v][v0];
      if (adj[v0][v] && nnei[v0][v]==3)
	{
	  ncontract[v]--;
	  ncontract[v0]--;
	}
    }

  nnei[v0][v2] = --nnei[v2][v0];
  nnei[v1][v3] = --nnei[v3][v1];
  if (adj[v1][v3] && nnei[v1][v3]==2)
    {
      ncontract[v1]++;
      ncontract[v3]++;
    }

  if (nnei[nv-1][v2] > 2)
    ncontract[v2]++;

  if (hole1)
    {
      //nnei[v0][v1] = ++nnei[v1][v0];
      nnei[v0][v2] = ++nnei[v2][v0];
    }

  if (hole2)
    {
      //nnei[v0][v3] = ++nnei[v3][v0];
      nnei[v0][v2] = ++nnei[v2][v0];
    }

  if (nnei[v0][v2] > 2)
    {
      ncontract[v0]--;
      ncontract[v2]--;
    }

  if (!adj[v1][v3])
    {
      if (!hole1) ncontract[v1]--;
      if (!hole2) ncontract[v3]--;
    }
  
  adj[v0][v2] = adj[v2][v0] = TRUE;

  nv--;
  ne -= 6;

  
  check_it(394);
  
}

/**********************************************************************/

static void
extendn_holes(int n, EDGE *e, EDGE *list[])

/* Deletes (n-3) edges in the next direction (e->next, e->next->next, ...)
   and their inverse and puts a valence
   n vertex into the resulting polygon.
   In list[0..(n-3)*2-1] the deleted edges are stored.
   This list must be handed to reducen() */
{
  register EDGE *work, *e1;
  int v[MAXN],i,j,k,vend;
  EDGE *oe, *owork;

  work = e;
  for (j=0;j<n-3;j++)
    {
      work = work->next;
      list[2*j] = work;
      list[2*j+1] = work->invers;
    }

  work = STARN(nv);
  firstedge[nv] = work+MAXN;
  firstedge[nv]->prev = work+MAXN+n-1;
  firstedge[nv]->prev->next = firstedge[nv];
  owork = firstedge[nv]->op;
  owork->next = firstedge[nv]->prev->op;
  owork->next->prev = owork;

  firstedge[e->start] = e;
  /* make sure firstedge points at a valid edge afterwards */

  /* work starts at the beginning of e: */

  work->start = work->invers->end = e->start;
  e1 = e;
  for (j=0;j<n-2;j++)
    e1 = e1->next;
  work->next = e1;
  work->next->prev = work;
  work->prev = e;
  e->next = work;
  oe = e->op;
  owork = work->op;
  owork->start = owork->invers->end = oe->start;
  owork->prev = e1->op;
  owork->prev->next = owork;
  owork->next = oe;
  oe->prev = owork;
  degree[e->start] = degree[e->start] - (n-4);

  v[0] = e->start;

  /* Now go one edge further around the pentagon: */
  e = e->invers->prev;

  work = work+1;
  work->start = work->invers->end = e->start;
  e->next->prev = work;
  work->next = e->next;
  work->prev = e;
  e->next = work;
  oe = e->op;
  owork = work->op;
  owork->start = owork->invers->end = oe->start;
  oe->prev->next = owork;
  owork->prev = oe->prev;
  owork->next = oe;
  oe->prev = owork;
  (degree[e->start])++;

  v[1] = e->start;

  /* Now go one edge further around the pentagon jumping over one edge: */

  for (j=2;j<n-1;j++)
    {
      e = list[2*j-3]->prev;

      firstedge[e->start] = e;
      /* here also an edge is deleted */

      work = work+1;
      work->start = work->invers->end = e->start;
      work->next = e->next->next;
      work->next->prev = work;
      work->prev = e;
      e->next = work;
      oe = e->op;
      owork = work->op;
      owork->start = owork->invers->end = oe->start;
      owork->prev = oe->prev->prev;
      owork->prev->next = owork;
      owork->next = oe;
      oe->prev = owork;
      /* the degree of e->start doesn't change */

      v[j] = e->start;
    }

  /* Finally go one edge further around the pentagon: */
  e = e1->invers;

  work = work+1;
  work->start = work->invers->end = e->start;
  e->next->prev = work;
  work->next = e->next;
  work->prev = e;
  e->next = work;
  oe = e->op;
  owork = work->op;
  owork->start = owork->invers->end = oe->start;
  oe->prev->next = owork;
  owork->prev = oe->prev;
  owork->next = oe;
  oe->prev = owork;
  (degree[e->start])++;

  v[n-1] = e->start;

  degree[nv] = n;

  /* do bookkeeping for removing edges (diagonals) */

  /* remove each diagonal edge.
     if nnei of its ends was 2 then it was contractible so reduce
     ncontract for its ends. */

  for (j=2;j<n-1;j++)
    {
      if (nnei[v[0]][v[j]] == 2)
	{
	  ncontract[v[0]]--;
	  ncontract[v[j]]--;
	}
      adj[v[0]][v[j]] = adj[v[j]][v[0]] = FALSE;
    }

  /* for each neighbor, vend, of v[0] reduce nnei of vend and
     each other end (v[j],j=2,..,n-2) of the diagonals.
     if vend and v[j] are adjacent and nnei is now 2 then increase
     ncontract for vend and v[j].
     don't bother with vend=(v[i],i=1,..,n-1) since nnei[v[i]][v[j]]
     would then need to be increased later
     when radial edges are added. (iou A) */

  e = list[2*(n-4)]->next;
  for (k=0;k<degree[v[0]]-3;k++)
    {
      e = e->next;
      vend = e->end;
      for (j=2;j<n-1;j++)
	{
	  nnei[v[j]][vend] = --nnei[vend][v[j]];
	  if (adj[v[j]][vend] && nnei[v[j]][vend]==2)
	    {
	      ncontract[vend]++;
	      ncontract[v[j]]++;
	    }
	}
    }

  /* for each neighbor, vend, of each non-v[0] end of the diagonals
     (v[j],j=2,..,n-2) reduce nnei of vend and v[0].
     if vend and v[0] are adjacent and nnei is now 2 then increase
     ncontract for vend and v[0].
     For holes we ignore the following comment to avoid checking for such 
     possible edge:
     don't bother with edges on path around v[nv] will deal with them later
     (iou B) */

  for (j=2;j<n-1;j++)
    {
      e = list[2*j-3];
      for (k=0;k<degree[v[j]]-1;k++)
	{
	  e = e->next;
	  vend = e->end;
	  nnei[v[0]][vend] = --nnei[vend][v[0]];
	  if (adj[v[0]][vend] && nnei[v[0]][vend]==2)
	    {
	      ncontract[vend]++;
	      ncontract[v[0]]++;
	    }
	}
    }

  /* do bookkeeping for adding edges (radial from new vertex) */

  /* for each old vertex, i, compute nnei[nv][i] by counting number of new
     neighbors of nv (v[j],j=0,..,n-1) which are adjacent to i. */

  for (i=0;i<nv;i++)
    {
      nnei[nv][i] = 0;
      for (j=0;j<n;j++)
	nnei[nv][i] += adj[v[j]][i];
      nnei[i][nv] = nnei[nv][i];
    }

  /* update nnei due to holes */

  e = firstedge[nv]->next;
  for (j=1;j<n-1;j++)
    {
      if (RIGHTHOLE(e)) 
	{ 
	  nnei[v[0]][v[j]] = --nnei[v[j]][v[0]];
	  nnei[nv][v[j]] = ++nnei[v[j]][nv];
	  nnei[v[0]][v[j+1]] = --nnei[v[j+1]][v[0]];
	  nnei[nv][v[j+1]] = ++nnei[v[j+1]][nv];
	  if (j==1 && nnei[v[0]][v[j]]==2)
	    {
	      ncontract[v[0]]++;
	      ncontract[v[j]]++;
	    }
	  else if (j+1==n-1 && nnei[v[0]][v[j+1]]==2)
	    {
	      ncontract[v[0]]++;
	      ncontract[v[j+1]]++;
	    }
	}
      e = e->next;
    }

  /* for each new edge if nnei is 2 increase ncontract of ends */

  ncontract[nv] = 0;
  for (j=0;j<n;j++)
    {
      adj[v[j]][nv] = adj[nv][v[j]] = TRUE;
      if (nnei[v[j]][nv] == 2)
	{
	  ncontract[v[j]]++;
	  ncontract[nv]++;
	}
    }

  /* update nnei[v[i]][v[j]] because of addition of nv.
     would need to increase for each pair (((i,j),i=0,..,n-2),j=i+1,..,n-1).
     from iou A above we didn't decrease (((i,j),i=1,n-2),j=i+1,..,n-1)
     except (1,n-1).
     The following is no longer true for holes:
     from iou B above we didn't decrease (((0,j),i=2,..,n-2),j=i-1,i+1).

     so still need to decrease ((0,j),j=3,..,n-3) and  increase (1,n-1).
     and increase (((0,j),i=2,..,n-2),j=i-1,i+1).
  */

  /* decrease ((0,j),j=3,..,n-3). */

  for (j=3;j<n-2;j++)
    {
      nnei[v[0]][v[j]] = --nnei[v[j]][v[0]];
      if (adj[v[0]][v[j]] && nnei[v[0]][v[j]]==2)
	{
	  ncontract[v[0]]++;
	  ncontract[v[j]]++;
	}
    }

  /* increase (1,n-1). */

  nnei[v[1]][v[n-1]] = ++nnei[v[n-1]][v[1]];
  if (adj[v[1]][v[n-1]] && nnei[v[1]][v[n-1]]==3)
    {
      ncontract[v[1]]--;
      ncontract[v[n-1]]--;
    }

  /* increase (((0,j),i=2,..,n-2),j=i-1,i+1). */

  for (i=2; i<n-1; i++)
    {
      nnei[v[0]][v[i-1]] = ++nnei[v[i-1]][v[0]];
      if (adj[v[0]][v[i-1]] && nnei[v[0]][v[i-1]]==3)
	{
	  ncontract[v[0]]--;
	  ncontract[v[i-1]]--;
	}
      nnei[v[0]][v[i+1]] = ++nnei[v[i+1]][v[0]];
      if (adj[v[0]][v[i+1]] && nnei[v[0]][v[i+1]]==3)
	{
	  ncontract[v[0]]--;
	  ncontract[v[i+1]]--;
	}
    }
  
  /* Now I have 6 edges and one vertex more */

  ne += 6;
  nv++;

  check_it(300+n);
}

/*****************************/

static void
reducen_holes(int n, EDGE *e, EDGE *list[])

/* The inverse operation to extendn().
   Deletes the vertex with valence n at the end of e->next. It is not
   checked whether the vertex really has valence n -- this has to be made
   sure in advance. The vector list[] must contain the edges deleted before.
   It might be that one of the edges leading to the new vertex now is
   an entry of firstedge[] */
{
  int v[MAXN],i,j,k,vend;
  EDGE *work;

  firstedge[e->start] = e;

  e->next = list[0];
  list[2*(n-4)]->next->prev = list[2*(n-4)];
  e->op->prev = list[0]->op;
  list[2*(n-4)]->op->prev->next = list[2*(n-4)]->op;
  degree[e->start] = degree[e->start] + (n-4);

  v[0] = e->start;

  e = e->invers;

  /* Delete the edge on the prev side */
  firstedge[e->start] = e;
  e->prev = e->prev->prev; e->prev->next = e;
  e->op->next = e->op->next->next; e->op->next->prev = e->op;
  (degree[e->start])--;

  v[1] = e->start;

  for (j=2;j<n-1;j++)
    {
      e = list[2*j-3]->next;
      firstedge[e->start] = e;
      /* Delete the edge and insert the old edge: */
      e->prev = list[2*j-3];
      list[2*j-3]->prev->next = list[2*j-3];
      e->op->next = list[2*j-3]->op;
      list[2*j-3]->op->next->prev = list[2*j-3]->op;

      v[j] = e->start;
    }

  e = list[2*n-8]->next->invers->next->next;
  firstedge[e->start] = e;
  e->prev = e->prev->prev; e->prev->next = e;
  e->op->next = e->op->next->next; e->op->next->prev = e->op;
  (degree[e->start])--;

  v[n-1] = e->start;

  work = STARN(nv-1)+MAXN;
  work->prev->next = STARN(nv-1) + MAXN + n;
  work->op->next->prev = work->prev->next->op;
  work->prev = STARN(nv-1) + MAXN + MAXN - 1;
  work->op->next = work->prev->op;

  /* do reverse operations as in extendn.
     see comments in extendn. */

  /* do bookkeeping for removing edges (radial from new vertex) */

  for (j=0;j<n;j++)
    {
      adj[v[j]][nv-1] = adj[nv-1][v[j]] = FALSE;
      if (nnei[v[j]][nv-1] == 2)
	ncontract[v[j]]--;
    }

  for (j=3;j<n-2;j++)
    nnei[v[0]][v[j]] = ++nnei[v[j]][v[0]];

  nnei[v[1]][v[n-1]] = --nnei[v[n-1]][v[1]];
  if (adj[v[1]][v[n-1]] && nnei[v[1]][v[n-1]] == 2)
    {
      ncontract[v[1]]++;
      ncontract[v[n-1]]++;
    }

  /* decrease (((0,j),i=2,..,n-2),j=i-1,i+1). */

  for (i=2; i<n-1; i++)
    {
      nnei[v[0]][v[i-1]] = --nnei[v[i-1]][v[0]];
      nnei[v[0]][v[i+1]] = --nnei[v[i+1]][v[0]];
    }
  if (nnei[v[0]][v[1]] == 2)
    {
      ncontract[v[0]]++;
      ncontract[v[1]]++;
    }
  if (nnei[v[0]][v[n-1]] == 2)
    {
      ncontract[v[0]]++;
      ncontract[v[n-1]]++;
    }
  
  /* update nnei due to holes */

  e = firstedge[v[0]];
  for (j=1;j<n-1;j++)
    {
      if (RIGHTHOLE(e)) 
	{ 
	  nnei[v[0]][v[j]] = ++nnei[v[j]][v[0]];
	  nnei[v[0]][v[j+1]] = ++nnei[v[j+1]][v[0]];
	  if (j==1 && nnei[v[0]][v[j]]==3)
	    {
	      ncontract[v[0]]--;
	      ncontract[v[1]]--;
	    }
	  else if (j+1 == n-1 && nnei[v[0]][v[j+1]]==3)
	    {
	      ncontract[v[0]]--;
	      ncontract[v[n-1]]--;
	    }
	}
      e = e->next;
    }

  /* do bookkeeping for adding edges (diagonals) */

  e = list[2*(n-4)]->next;
  for (k=0;k<degree[v[0]]-(n-1);k++)
    {
      e = e->next;
      vend = e->end;
      for (j=2;j<n-1;j++)
	{
	  nnei[v[j]][vend] = ++nnei[vend][v[j]];
	  if (adj[v[j]][vend] && nnei[v[j]][vend]==3)
	    {
	      ncontract[vend]--;
	      ncontract[v[j]]--;
	    }
	}
    }

  for (j=2;j<n-1;j++)
    {
      e = list[2*j-3];
      for (k=0;k<degree[v[j]]-1;k++)
	{
	  e = e->next;
	  vend = e->end;
	  nnei[v[0]][vend] = ++nnei[vend][v[0]];
	  if (adj[v[0]][vend] && nnei[v[0]][vend]==3)
	    {
	      ncontract[vend]--;
	      ncontract[v[0]]--;
	    }
	}
    }

  for (j=2;j<n-1;j++)
   adj[v[0]][v[j]] = adj[v[j]][v[0]] = TRUE;

  for (j=2;j<n-1;j++)
    if (nnei[v[0]][v[j]] == 2)
      {
	ncontract[v[0]]++;
	ncontract[v[j]]++;
      }

  nv--;
  ne-=6;

  
  check_it(390+n);
  
}

/**************************************************************************/

static void
extend_four(EDGE *ref)

/* This routine is another implementation of the 4-operation
   in Batagelj's paper.
*/
{
    register EDGE *a,*b,*start,*dummy;
    EDGE *oa,*ob,*ostart,*odummy;
    int v0,v1,v2,v3,i,k,v;
    EDGE *list0,*list1,*e,*oref;

    list0 = ref->next;
    list1 = ref->invers->next;

    start=four_op(nv);
    ostart=start+NUMEDGES;
    b=ref->next->invers;
    ob=b->op;
    a=b->next;
    oa=a->op;
    oref=ref->op;

    degree[nv]=4; firstedge[nv]=start;

    dummy=start+1;
    odummy=ostart+1;
    v3 = a->start;
    start->end=dummy->start=v3;
    ostart->end=odummy->start=v3;
    (degree[v3])++;
    a->prev=b->next=dummy;
    oa->next=ob->prev=odummy;
    dummy->prev=b; dummy->next=a;
    odummy->next=ob; odummy->prev=oa;
    adj[v3][nv] = adj[nv][v3] = TRUE;

    b=b->invers; a=ref->prev;
    ob=ob->invers; oa=oref->next;

    start+=2; /*2*/ dummy=start+1;
    ostart+=2;      odummy=ostart+1;
    v0 = b->start;
    start->end=dummy->start=v0;
    ostart->end=odummy->start=v0;
    firstedge[b->start]=dummy;
    b->prev=a->next=dummy;
    ob->next=oa->prev=odummy;
    dummy->prev=a; dummy->next=b;
    odummy->next=oa; odummy->prev=ob;
    adj[v0][nv] = adj[nv][v0] = TRUE;

    a=a->invers; b=a->prev;
    oa=oa->invers; ob=oa->next;
    start+=2; /*4*/ dummy=start+1;
    ostart+=2;      odummy=ostart+1;
    v1 = b->start;
    start->end=dummy->start=v1;
    ostart->end=odummy->start=v1;
    (degree[v1])++;
    a->prev=b->next=dummy;
    oa->next=ob->prev=odummy;
    dummy->prev=b; dummy->next=a;
    odummy->next=ob; odummy->prev=oa;
    adj[v1][nv] = adj[nv][v1] = TRUE;

    start->next=(start-4)->prev=ref;
    ostart->prev=(ostart-4)->next=oref;

    ref->next=start-4; ref->prev=start;
    oref->prev=ostart-4; oref->next=ostart;
    ref->start=ref->invers->end=nv;
    oref->start=oref->invers->end=nv;
    v2 = ref->end;
    adj[v2][nv] = adj[nv][v2] = TRUE;

    adj[v0][v2] = adj[v2][v0] = FALSE;

  for (i=0;i<nv;i++)
    nnei[i][nv] = nnei[nv][i] =
      adj[v0][i] + adj[v1][i] + adj[v2][i] + adj[v3][i];

  if (nnei[v0][v2] > 2)
    {
      ncontract[v0]++;
      ncontract[v2]++;
    }

  if (adj[v1][v3])
    ncontract[nv] = 2;
  else
    {
      ncontract[nv] = 4;
      ncontract[v1]++;
      ncontract[v3]++;
    }

  nnei[v0][v2] = ++nnei[v2][v0];

  nnei[v1][v3] = ++nnei[v3][v1];

  e = list0;
  for (k=0;k<degree[v0]-3;k++)
    {
      e = e->next;
      v = e->end;
      nnei[v2][v] = --nnei[v][v2];
      if (adj[v2][v] && nnei[v2][v]==2)
	{
	  ncontract[v]++;
	  ncontract[v2]++;
	}
    }

  e = list1;
  for (k=0;k<degree[v2]-3;k++)
    {
      e = e->next;
      v = e->end;
      nnei[v0][v] = --nnei[v][v0];
      if (adj[v0][v] && nnei[v0][v]==2)
	{
	  ncontract[v]++;
	  ncontract[v0]++;
	}
    }
  
    nv++;
    ne+=6;

    check_it(804);
}

/**************************************************************************/

static int
is_min4_four(EDGE *ref)

/* Test if ref (known to be an edge!) is the reference edge of a
   legal four-reduction in the min4 case. 
   Return 1 if it is, and 0 if it isn't.
*/
{
    int w;

    if (degree[ref->start] != 4) return 0;
    if (degree[ref->prev->end] == 4 || degree[ref->next->end] == 4) return 0;

    w = ref->next->next->end;
    if (adj[w][ref->end])
      return 0;
    else
      return 1;

}

/**************************************************************************/

static void
reduce_four(EDGE *ref)

/* The reverse of extend_four(ref) */
{
    register EDGE *a,*b;
    EDGE *oa,*ob,*oref,*e;
    int v,v0,v1,v2,v3,i;

    a=ref->invers->prev->invers;
    b=a->prev->prev;
    oa=a->op;
    ob=b->op;
    oref=ref->op;

    a->prev=b; b->next=a; 
    oa->next=ob; ob->prev=oa;
    v3=a->start;
    firstedge[v3]=a; (degree[v3])--;
    adj[v3][nv-1]=adj[nv-1][v3]=FALSE;

    a=b->invers; b=a->prev->prev;
    oa=a->op;
    ob=b->op;
    a->prev=b->next=ref;
    oa->next=ob->prev=oref;
    ref->next=a; ref->prev=b;
    oref->prev=oa; oref->next=ob;
    v0=a->start;
    ref->start=ref->invers->end=v0;
    oref->start=oref->invers->end=v0;
    firstedge[v0]=a;
    adj[v0][nv-1]=adj[nv-1][v0]=FALSE;

    a=b->invers; b=a->prev->prev;
    oa=a->op;
    ob=b->op;
    a->prev=b; b->next=a; 
    oa->next=ob; ob->prev=oa;
    v1=a->start;
    firstedge[v1]=a; (degree[v1])--;
    adj[v1][nv-1]=adj[nv-1][v1]=FALSE;

    v2=ref->end;
    adj[v2][nv-1]=adj[nv-1][v2]=FALSE;

  e = ref->next;
  for (i=0;i<degree[v0]-3;i++)
    {
      e = e->next;
      v = e->end;
      nnei[v2][v] = ++nnei[v][v2];
      if (adj[v2][v] && nnei[v2][v]==3)
	{	
	  ncontract[v]--;
	  ncontract[v2]--;
	}
    }
  e = ref->invers->next;
  for (i=0;i<degree[v2]-3;i++)
    {
      e = e->next;
      v = e->end;
      nnei[v0][v] = ++nnei[v][v0];
      if (adj[v0][v] && nnei[v0][v]==3)
	{
	  ncontract[v]--;
	  ncontract[v0]--;
	}
    }

  nnei[v0][v2] = --nnei[v2][v0];
  nnei[v1][v3] = --nnei[v3][v1];

  if (nnei[v0][v2] > 2)
    {
      ncontract[v0]--;
      ncontract[v2]--;
    }

  if (!adj[v1][v3])
    {
      ncontract[v1]--;
      ncontract[v3]--;
    }

  adj[v0][v2] = adj[v2][v0] = TRUE;

    nv--;
    ne-=6;

    check_it(894);
}

